/*															*/
/* AtMega328p/@16MHz USART 9600BAUD							*/
/*															*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include <stdio.h>
#include "FatFs/ff.h"
#include "MMC/mmc.h"

#define FIFO_SIZE 16				/* FIFO buffer size		*/
#define BAUD 9600					/* USART BAUD			*/

/* USART PORTS define	*/
#define UBRR  UBRR0
#define UCSRB UCSR0B
#define UDR   UDR0
#define RXEN  RXEN0
#define TXEN  TXEN0
#define RXCIE RXCIE0
#define UDRIE UDRIE0

/* USART interrupt control define	*/
#define USART_RECV_ENABLE()  ( UCSRB |=  _BV(RXCIE) )
#define USART_RECV_DISABLE() ( UCSRB &= ~_BV(RXCIE) )

#define USART_SEND_ENABLE()  ( UCSRB |=  _BV(UDRIE) )
#define USART_SEND_DISABLE() ( UCSRB &= ~_BV(UDRIE) )

typedef struct FIFO_t {				/* FIFO buffer struct	*/
	uint8_t idx_w;
	uint8_t idx_r;
	uint8_t count;
	uint8_t buff[FIFO_SIZE];
} FIFO;

static volatile FIFO txfifo;		/* send FIFO buffer 	*/
static volatile FIFO rxfifo;		/* recv FIFO buffer 	*/

/* Timer1 10msec Interval timer			*/
ISR( TIMER1_COMPA_vect )
{
	disk_timerproc();	/* MMC Timer	*/
}

/* UART RECV interrupt					*/
ISR( USART_RX_vect )
{
	uint8_t data = UDR;
	if ( rxfifo.count < FIFO_SIZE ) {
		rxfifo.count++;
		rxfifo.buff[ rxfifo.idx_w++ ] = data;
		if ( rxfifo.idx_w >= FIFO_SIZE ) rxfifo.idx_w = 0;
	}
}

/* UART SEND registr empty interrupt	*/
ISR( USART_UDRE_vect )
{
	UDR = txfifo.buff[ txfifo.idx_r++ ];
	if ( txfifo.idx_r >= FIFO_SIZE ) txfifo.idx_r = 0;
	if ( --txfifo.count == 0 ) USART_SEND_DISABLE();
}

/* USART Get recv data count			*/
uint8_t usart_recv_test( void )
{
	return ( rxfifo.count );
}

/* USART Get recv data byte				*/
uint8_t usart_getchar( void )
{
	uint8_t data;
	while ( !usart_recv_test() );
	USART_RECV_DISABLE();
	data = rxfifo.buff[ rxfifo.idx_r++ ];
	if ( rxfifo.idx_r >= FIFO_SIZE ) rxfifo.idx_r = 0;
	rxfifo.count--;
	USART_RECV_ENABLE();
	return ( data );
}

/* USART Get send buffer free size		*/
uint8_t usart_send_test( void )
{
	return ( FIFO_SIZE - txfifo.count );	/* calc free size */
}

/* USART Put send byte data				*/
void usart_putchar( char data )
{
	while ( !usart_send_test() );
	USART_SEND_DISABLE();
	txfifo.count++;
	txfifo.buff[ txfifo.idx_w++ ] = data;
	if ( txfifo.idx_w >= FIFO_SIZE ) txfifo.idx_w = 0;
	USART_SEND_ENABLE();
}

/* stdout setup							*/
static int _usart_putchar( char c ,FILE *stream )
{
	usart_putchar( c );
	if ( c == '\n' ) {
		usart_putchar( '\r' );
	}
	return ( 0 );
}
static FILE my_stdout = FDEV_SETUP_STREAM( _usart_putchar ,NULL ,_FDEV_SETUP_WRITE );

/* AVR Initialization					*/
static void hardwareInit( void )
{
	/* PortB  +--------7: Pullup : non use  */
	/*        |+-------6: Pullup : non use  */
	/*        ||+------5: HIZ    : SD SCK   */
	/*        |||+-----4: HIZ    : SD MISO  */
	/*        ||||+----3: HIZ    : SD MOSI  */
	/*        |||||+---2: Out(H) : SD #CS   */
	/*        ||||||+--1: Pullup : non use  */
	/*        |||||||+-0: Pullup : non use	*/
	DDRB  = 0b00000100;
	PORTB = 0b11000111;

	/* PortC  +--------7: Pullup : non use 	*/
	/*        |+-------6: Pullup : non use 	*/
	/*        ||+------5: Pullup : non use 	*/
	/*        |||+-----4: Pullup : non use 	*/
	/*        ||||+----3: Pullup : non use 	*/
	/*        |||||+---2: Pullup : non use 	*/
	/*        ||||||+--1: Pullup : non use	*/
	/*        |||||||+-0: Out(L) : SD #POW  */
	DDRC  = 0b00000001;
	PORTC = 0b11111110;

	/* PortD  +--------7: Pullup : non use	*/
	/*        |+-------6: Pullup : non use	*/
	/*        ||+------5: Pullup : non use	*/
	/*        |||+-----4: Pullup : non use	*/
	/*        ||||+----3: Pullup : non use	*/
	/*        |||||+---2: Pullup : non use  */
	/*        ||||||+--1: Out(L) : TXD      */
	/*        |||||||+-0: Pullup : RXD      */
	DDRD  = 0b00000010;
	PORTD = 0b11111101;

	/* Timer1 setup							*/
	OCR1A = (uint16_t)( ( F_CPU / 256L ) / 100L );
	TCNT1 = 0;
	TCCR1B = 0x04 | _BV(WGM12);		/* F_CPU / 256 : CTC	*/
	TIMSK1 |= _BV(OCIE1A);

	/* UART Initialization					*/
	memset( (void *)&txfifo ,0x00 ,FIFO_SIZE );	/* TX FIFO buffer clear		*/
	memset( (void *)&rxfifo ,0x00 ,FIFO_SIZE );	/* RX FIFO buffer clear		*/
	UBRR = (uint16_t)( F_CPU / ( 16UL * (uint32_t)BAUD ) ) - 1;	/* set BAUD */
	UCSRB = _BV(RXEN) | _BV(TXEN);			/* RX / TX ENABLE				*/
	USART_RECV_ENABLE();					/* USART recv interrupt enable	*/

	stdout = &my_stdout;					/* Stdout setup					*/

	SMCR = _BV(SE); 						/* Power seve enabel for IDOL 	*/

	sei();		  							/* enable global interrupt 		*/
}

/* MAIN						*/
int main( void )
{
	FRESULT res;
	FATFS fs;
	DIR dir;
	FILINFO info;
	FIL fil;
	uint16_t cnt;
	char buf[64];
	int ret;
	
	hardwareInit();

	while ( 1 ) {

		/* Enter key wait	*/
		asm( "sleep" );									/* CPU IDOL MODE	*/
		if ( ( usart_recv_test() == 0 ) || ( usart_getchar() != '\r' ) ) {
			continue;
		}

		/* Listup files		*/
		f_mount( 0 ,&fs );								/* re mount			*/

		strcpy( buf ,"<-------------------------------------------------->\n" );
		printf( "%s" ,buf );

		res = f_opendir( &dir ,"" );
		if ( res != FR_OK ) {
			printf( "f_opendir() : error %u\n\n" ,res );
			continue;
		}

		res = f_open( &fil ,"test.txt", FA_CREATE_ALWAYS | FA_WRITE);
		if ( res != FR_OK ) {
			printf( "f_open() : error %u\n\n" ,res );
			continue;
		}

		cnt = 0;
		while ( 1 ) {
			res = f_readdir( &dir ,&info );
			if ( ( res != FR_OK ) || ( info.fname[0] == 0 ) ) {
				break;
			}

			if ( info.fname[0] == '.') {
				continue;
			}

            if ( info.fattrib & AM_DIR ) {
				strcpy( buf ,"[DIR] : " );
			} else {
				strcpy( buf ,"[FIL] : " );
			}

			strcat( buf ,info.fname );
			strcat( buf ,"\n" );
			printf( "%s" ,buf );

			ret = f_puts( buf ,&fil );
			if ( ret == EOF ) {
				f_close( &fil );
				printf( "f_puts() : error\n\n" );
				continue;
			}

			cnt++;
       	}

		sprintf( buf ,"%uFiles\n\n" ,cnt );
		printf( "%s" ,buf );

		ret = f_puts( buf ,&fil );
		if ( ret == EOF ) {
			printf( "f_puts() : error\n\n" );
		} else {
			printf( "'test.txt' is create OK !\n\n" );
		}

		f_close( &fil );
	}

	return ( 0 );
}
